﻿//----------------------------------------------------
// InsertTitle.cs (c) 2019 by Enes Siljak
//----------------------------------------------------

using System;
using System.Linq;
using System.IO;
using System.Collections.Generic;
using Autodesk.AutoCAD.Runtime;
using Autodesk.AutoCAD.ApplicationServices;
using Autodesk.AutoCAD.DatabaseServices;
using Autodesk.AutoCAD.Geometry;
using Autodesk.AutoCAD.EditorInput;
using Autodesk.AutoCAD.Windows;

namespace XenAlati
{
    public class InsertTitle
    {
        string path;
        string fBlockDef;
        string blockName;
        object[] numericCounters;
        int countLayouts = 0;
        Dictionary<string, string> attributesDictionary = new Dictionary<string, string>();
        
        [CommandMethod("XEN_INSERTTITLE")]
        public void Xen_InsertTitle()
        {
            Editor acDocEd = Application.DocumentManager.MdiActiveDocument.Editor;

            countLayouts = 0;
            attributesDictionary.Clear();
            path = null;
            fBlockDef = null;
            blockName = null;
            numericCounters = null;

            //folder with files to process
            using (var fldrDlg = new System.Windows.Forms.FolderBrowserDialog())
            {
                fldrDlg.Description = "Select Folder With Files To Insert/Update Title Block";
                fldrDlg.ShowNewFolderButton = false;
                fldrDlg.SelectedPath = Path.GetDirectoryName(fBlockDef);
                if (fldrDlg.ShowDialog() == System.Windows.Forms.DialogResult.OK)
                {
                    path = fldrDlg.SelectedPath;
                }
                else return;
            }
            if (path == null)
                return;
            path = PathAddBackslash(path);

            //get file with title block
            fBlockDef = SelectDwgFileWithBlocktDef();
            if (fBlockDef == null)
                return;
            acDocEd.WriteMessage(string.Format("\nSelected File: {0}\n", fBlockDef));

            //get block
            blockName = SelectBlock();
            if (blockName == null)
                return;
            acDocEd.WriteMessage(string.Format("Selected Block: {0}\n", blockName));

            //setup attributes
            string textdefatt = ListBlockAttributes();
            if (textdefatt == null)
                return;
            ParseAttributes(textdefatt);
            
                        

            Database acSourceDb = new Database(false, true);
            acSourceDb.ReadDwgFile(fBlockDef, FileOpenMode.OpenForReadAndAllShare, true, "");
            using (Transaction acTransSource = acSourceDb.TransactionManager.StartTransaction())
            {
                try
                {
                    BlockTable blkTable = (BlockTable)acTransSource.GetObject(acSourceDb.BlockTableId, OpenMode.ForRead);
                    ObjectId blkRecId = blkTable[blockName];
                    
                    Point3d insPoint = GetInsertionPoint(acSourceDb, acTransSource);

                    ObjectIdCollection acObjIdColl = new ObjectIdCollection();
                    acObjIdColl.Add(blkRecId);
                    
                    DirectoryInfo d = new DirectoryInfo(path);
                    FileInfo[] Files = d.GetFiles("*.dwg");
                    IList<FileInfo> sortedFiles = GetOrderedFiles(Files);
                    
                    foreach (FileInfo file in sortedFiles)
                    {
                        if (string.Compare(fBlockDef, file.FullName) == 0)
                            continue;

                        acDocEd.WriteMessage(string.Format("\n{0}", file.FullName));
                        System.Windows.Forms.Application.DoEvents();


                        using (Database acDestDb = new Database(false, true))
                        {
                            acDestDb.ReadDwgFile(file.FullName, FileOpenMode.OpenForReadAndAllShare, false, null);
                            using (Transaction acTransDest = acDestDb.TransactionManager.StartTransaction())
                            {
                                BlockTable acBlkTblDest = acTransDest.GetObject(acDestDb.BlockTableId, OpenMode.ForRead) as BlockTable;
                                if (!acBlkTblDest.Has(blockName))
                                {
                                    IdMapping mapping = new IdMapping();
                                    acSourceDb.WblockCloneObjects(acObjIdColl, acDestDb.BlockTableId, mapping, DuplicateRecordCloning.Replace, false);

                                    ObjectId blkDestRecId = mapping[blkRecId].Value;
                                    InsertRefInLayouts(blkDestRecId, acTransDest, acDestDb, insPoint);
                                }
                                else
                                {
                                    UpdateRefInLayouts(acTransDest, acDestDb);
                                }
                                
                                acTransDest.Commit();
                            }

                            acDestDb.SaveAs(file.FullName, DwgVersion.Current);
                            acDestDb.Dispose();
                        }
                    }
                    acDocEd.WriteMessage(string.Format("\nProcessed {0} layouts.\n", countLayouts));
                }
                catch (System.Exception ex)
                {
                    acDocEd.WriteMessage("\n" + ex.ToString());
                }

                acTransSource.Commit();
            }
            acSourceDb.Dispose();
            
        }

        void InsertRefInLayouts(ObjectId blkRecId, Transaction acTrans, Database acCurDb, Point3d insPoint)
        {
            Editor acDocEd = Application.DocumentManager.MdiActiveDocument.Editor;

            if (blkRecId != ObjectId.Null)
            {
                BlockTableRecord acBlkTblRec = acTrans.GetObject(blkRecId, OpenMode.ForRead) as BlockTableRecord;
                DBDictionary layouts = acTrans.GetObject(acCurDb.LayoutDictionaryId, OpenMode.ForRead) as DBDictionary;

                foreach (DBDictionaryEntry entry in layouts)
                {
                    if (entry.Key == "Model")
                        continue;
                    
                    Layout layEx = (Layout)acTrans.GetObject(entry.Value, OpenMode.ForRead);
                    BlockTableRecord blkBlkRec = acTrans.GetObject(layEx.BlockTableRecordId, OpenMode.ForWrite) as BlockTableRecord;

                    acDocEd.WriteMessage(", " + layEx.LayoutName);
                    System.Windows.Forms.Application.DoEvents();

                    using (BlockReference acBlkRef = new BlockReference(insPoint, blkRecId))
                    {
                        blkBlkRec.AppendEntity(acBlkRef);
                        acTrans.AddNewlyCreatedDBObject(acBlkRef, true);
                        
                        if (acBlkTblRec.HasAttributeDefinitions)
                        {
                            foreach (ObjectId objID in acBlkTblRec)
                            {
                                DBObject dbObj = acTrans.GetObject(objID, OpenMode.ForRead) as DBObject;
                                if (dbObj is AttributeDefinition)
                                {
                                    AttributeDefinition acAtt = dbObj as AttributeDefinition;
                                    if (!acAtt.Constant)
                                    {
                                        using (AttributeReference acAttRef = new AttributeReference())
                                        {
                                            acAttRef.SetAttributeFromBlock(acAtt, acBlkRef.BlockTransform);
                                            acAttRef.Position = acAtt.Position.TransformBy(acBlkRef.BlockTransform);
                                            acAttRef.AdjustAlignment(acCurDb);
                                            acBlkRef.AttributeCollection.AppendAttribute(acAttRef);
                                            
                                            try
                                            {
                                                string value = attributesDictionary[acAtt.Tag.ToUpper()];
                                                acAttRef.TextString = string.Format(attributesDictionary[acAtt.Tag.ToUpper()], numericCounters);
                                            }
                                            catch (KeyNotFoundException)
                                            {
                                            }
                                            
                                            acTrans.AddNewlyCreatedDBObject(acAttRef, true);
                                        }
                                    }
                                }
                            }
                        }
                    }
                    IncCounters();
                }
            }
        }
        
        void UpdateRefInLayouts(Transaction acTrans, Database acCurDb)
        {
            Editor acDocEd = Application.DocumentManager.MdiActiveDocument.Editor;

            DBDictionary layouts = acTrans.GetObject(acCurDb.LayoutDictionaryId, OpenMode.ForRead) as DBDictionary;

            foreach (DBDictionaryEntry entry in layouts)
            {
                if (entry.Key == "Model")
                    continue;
                
                Layout layEx = (Layout)acTrans.GetObject(entry.Value, OpenMode.ForRead);
                BlockTableRecord blkBlkRec = acTrans.GetObject(layEx.BlockTableRecordId, OpenMode.ForWrite) as BlockTableRecord;

                acDocEd.WriteMessage(", " + layEx.LayoutName);
                System.Windows.Forms.Application.DoEvents();

                foreach (ObjectId entId in blkBlkRec)
                {
                    Entity ent = acTrans.GetObject(entId, OpenMode.ForRead) as Entity;
                    if (ent != null)
                    {
                        BlockReference br = ent as BlockReference;
                        if (br != null)
                        {
                            BlockTableRecord bd = (BlockTableRecord)acTrans.GetObject(br.BlockTableRecord, OpenMode.ForRead);
                            
                            if (bd.Name.ToUpper() == blockName.ToUpper())
                            {
                                foreach (ObjectId arId in br.AttributeCollection)
                                {
                                    DBObject obj = acTrans.GetObject(arId, OpenMode.ForRead);
                                    AttributeReference ar = obj as AttributeReference;
                                    if (ar != null)
                                    {
                                        try
                                        {
                                            string attValue = attributesDictionary[ar.Tag.ToUpper()];
                                            ar.UpgradeOpen();
                                            ar.TextString = string.Format(attValue, numericCounters);
                                            ar.DowngradeOpen();
                                        }
                                        catch (KeyNotFoundException)
                                        {
                                        }
                                        
                                    }
                                }
                            }
                        }
                    }
                }
                IncCounters();
            }
        }
               
        Point3d GetInsertionPoint(Database db, Transaction tr)
        {
            Point3d result = new Point3d(0, 0, 0);
            DBDictionary layoutDict = tr.GetObject(db.LayoutDictionaryId, OpenMode.ForRead) as DBDictionary;
            foreach (DBDictionaryEntry entry in layoutDict)
            {
                if (entry.Key == "Model") continue;

                var layout = (Layout)tr.GetObject(entry.Value, OpenMode.ForRead);
                var layoutBtr = (BlockTableRecord)tr.GetObject(layout.BlockTableRecordId, OpenMode.ForRead);

                bool found = false;
                foreach (ObjectId id in layoutBtr)
                {
                    if (id.ObjectClass.DxfName == "INSERT")
                    {
                        var br = (BlockReference)tr.GetObject(id, OpenMode.ForRead);
                        if (br.Name.Equals(blockName, StringComparison.CurrentCultureIgnoreCase))
                        {
                            result = br.Position;
                            found = true;
                            break;
                        }
                    }
                }
                if (found)
                    break;
            }

            return result;
        }

        void ParseAttributes(string textdefatt)
        {

            string attributes = textdefatt;
            string counters = null;
            
            int ic = textdefatt.LastIndexOf('!');

            if (ic > 0)
            {
                attributes = Left(textdefatt, ic);
                counters = textdefatt.Substring(ic + 1);
            }
            
            string[] stringSeparators = new string[] { "\r\n" };
            string[] attlines = attributes.Split(stringSeparators, StringSplitOptions.RemoveEmptyEntries);
            for (int j = 0; j < attlines.Length; j++)
            {
                attlines[j] = attlines[j].Trim();

                int idx = attlines[j].IndexOf(' ');
                string attTag = Left(attlines[j], idx).Trim();
                string attValue = attlines[j].Substring(idx + 1).Trim();
                if(attValue != "")
                {
                    if (attValue == "!-")
                    {
                        attValue = "";
                    }
                    try
                    {
                        attributesDictionary.Add(attTag.ToUpper(), attValue);
                    }
                    catch (ArgumentException)
                    {
                    }
                   
                }
                
            }



            string[] counterlines = null;
            if (counters != null)
            {
                counterlines = counters.Split(stringSeparators, StringSplitOptions.RemoveEmptyEntries);
                numericCounters = new object[counterlines.Length];
            }

            for (int k = 0; k < counterlines.Length; k++)
            {
                if (counterlines[k].Trim() != "")
                {
                    string[] cntr = counterlines[k].Split(':');
                    numericCounters[k] = int.Parse(cntr[1]);
                }
            }
        }
        
        string SelectDwgFileWithBlocktDef()
        {
            OpenFileDialog ofd = new OpenFileDialog(
                "Select Dwg with Title Block to insert",
                path,
                "dwg",
                "Dwg With Title Block To Insert",
                OpenFileDialog.OpenFileDialogFlags.ForceDefaultFolder | OpenFileDialog.OpenFileDialogFlags.DefaultIsFolder
              );

            System.Windows.Forms.DialogResult dr = ofd.ShowDialog();
            if (dr != System.Windows.Forms.DialogResult.OK)
                return null;

            return ofd.Filename;
        }

        string SelectBlock()
        {
            Database acExDb = new Database(false, true);
            acExDb.ReadDwgFile(fBlockDef, FileOpenMode.OpenForReadAndAllShare, true, "");
            List<string> blocks = new List<string>();
            using (Transaction acTransEx = acExDb.TransactionManager.StartTransaction())
            {
                BlockTable blkTable = (BlockTable)acTransEx.GetObject(acExDb.BlockTableId, OpenMode.ForRead);
                foreach (ObjectId id in blkTable)
                {
                    BlockTableRecord btRecord = (BlockTableRecord)acTransEx.GetObject(id, OpenMode.ForRead);
                    if (!btRecord.IsLayout)
                    {
                        blocks.Add(btRecord.Name);
                        //Application.DocumentManager.MdiActiveDocument.Editor.WriteMessage(string.Format("\nBlock name: {0}", btRecord.Name));
                    }
                }
                acTransEx.Abort();
            }
            acExDb.Dispose();

            if(blocks.Count == 0)
            {
                Application.DocumentManager.MdiActiveDocument.Editor.WriteMessage("Block not found in selected drawing!");
                return null;
            }

            string result = "";
            SimpleDlg.ShowSelectOptionDialog("Select Block", "Blocks In File:", blocks, ref result);
            return result;
        }

        string ListBlockAttributes()
        {
            Database acExDb = new Database(false, true);
            acExDb.ReadDwgFile(fBlockDef, FileOpenMode.OpenForReadAndAllShare, true, "");
            List<string> blockAttributes = new List<string>();
            using (Transaction acTransEx = acExDb.TransactionManager.StartTransaction())
            {
                BlockTable blkTable = (BlockTable)acTransEx.GetObject(acExDb.BlockTableId, OpenMode.ForRead);

                ObjectId blkRecId = blkTable[blockName];
                BlockTableRecord acBlkTblRec = (BlockTableRecord)acTransEx.GetObject(blkRecId, OpenMode.ForRead);

                foreach (ObjectId objID in acBlkTblRec)
                {
                    DBObject dbObj = acTransEx.GetObject(objID, OpenMode.ForRead) as DBObject;

                    if (dbObj is AttributeDefinition)
                    {
                        AttributeDefinition acAtt = dbObj as AttributeDefinition;
                        blockAttributes.Add( string.Format( "/* {0,-40} */ {1,-20}"  , acAtt.Prompt , acAtt.Tag ));

                        //Application.DocumentManager.MdiActiveDocument.Editor.WriteMessage("\nATTR TAG: " + acAtt.Tag);
                    }

                }
                    acTransEx.Abort();
            }
            acExDb.Dispose();

            blockAttributes.Sort();
            return EditTitleBlockDialog.Show(blockAttributes);
        }

        void IncCounters()
        {
            countLayouts++;

            if (numericCounters != null)
            {
                int incval;
                for (int i = 0; i < numericCounters.Length; i++)
                {
                    incval = (int)numericCounters[i];
                    incval++;
                    numericCounters[i] = (object)incval;
                }
            }
            
        }

        IList<FileInfo> GetOrderedFiles(FileInfo[] files)
        {
            List<FileInfo> fileList = new List<FileInfo>(files);
            int compare(FileInfo file1, FileInfo file2)
            {
                return file1.Name.CompareTo(file2.Name);
            }

            fileList.Sort(compare);
            return fileList;
        }

        string PathAddBackslash(string path)
        {
            if (path == null)
                throw new ArgumentNullException(nameof(path));

            path = path.TrimEnd();

            if (PathEndsWithDirectorySeparator())
                return path;

            return path + GetDirectorySeparatorUsedInPath();

            bool PathEndsWithDirectorySeparator()
            {
                if (path.Length == 0)
                    return false;

                char lastChar = path[path.Length - 1];
                return lastChar == Path.DirectorySeparatorChar
                    || lastChar == Path.AltDirectorySeparatorChar;
            }

            char GetDirectorySeparatorUsedInPath()
            {
                if (path.Contains(Path.AltDirectorySeparatorChar))
                    return Path.AltDirectorySeparatorChar;

                return Path.DirectorySeparatorChar;
            }
        }

        public static string Left(string value, int maxLength)
        {
            if (string.IsNullOrEmpty(value)) return value;
            maxLength = Math.Abs(maxLength);

            return (value.Length <= maxLength
                   ? value
                   : value.Substring(0, maxLength)
                   );
        }

    }

}
